#ifndef __CPU_H
#define __CPU_H 1


#include <stdio.h>
#include <limits.h>
#include <time.h>
#include <math.h>

#include "general.h"
#include "mem.h"
#include "gcube.h"

// registers
extern __u32 regs[4096];
extern __u64 fpregs[32];


#define BUS_CLOCK_SPEED					0x09a7ec80
// 486 MHz
#define CPU_CLOCK_SPEED					0x1cf7c580

#define TBMAGIC									((__u64) 23 * 365 * 24 * 60 * 60 + 7 * 366 * 24 * 60 * 60)
#define T2TICKS(t)							((t) * CPU_CLOCK_SPEED / 12)
#define CALC_TBL								(TBL = T2TICKS ((__u64) time (NULL) - TBMAGIC))
#define CALC_TBU								(TBU = T2TICKS ((__u64) time (NULL) - TBMAGIC) >> 32)


// opcodes
#define OPCODE(X)			 		static void X (__u32 opcode)
#define OP4(X,O,I,R)			(op##X [(O << 10) | (I << 1) | R])
#define OP3(X,I,R)				(op##X [(I << 1) | R])
#define OP1(X)						(op0 [X])
#define OPR(X,I,R,m,n,o)\
	({\
		int i;\
		for (i = 0; i < (1 << (n - m)); i++)\
			op##X [(i << m) | (I << 1) | R] = o;\
	})

// bit operations (rewrite in asm)
#define BIT(X,m,n)				(((X) >> ((m)-(n)-1)) & 1)
#define BIT_IS_SET(X,m,n)	((X) & (1 << ((m)-(n)-1)))
#define BIT32(X,n)				(BIT (X, 32, n))
#define SET_BIT(X,n)			((X) |= (1 << (31-(n))))
#define CLEAR_BIT(X,n)		((X) &= ~(1 << (31-(n))))
#define CHANGE_BIT(X,n,b)	((X) = ((X) &~ (1 << (31-(n)))) | ((b) << (31-(n))))


#define REGS(x)	(regs[x])
// general purpose registers (r0-r31)
#define I_GPR		0
#define GPR 		regs
// segment registers (sr0-sr15)
#define I_SR		32
#define SR			(&regs[I_SR])
// condition register
#define I_CR		48
#define CR			(regs[I_CR])
// machine state register
#define I_MSR		49
#define MSR			(regs[I_MSR])
// floating point status and control register
#define I_FPSCR	50
#define FPSCR		(regs[I_FPSCR])
// time base facility (lower and upper)
#define I_TBL		51
#define I_TBU		52
#define TBL			(regs[I_TBL])
#define TBU			(regs[I_TBU])
// special purpose registers (spr0-spr1023)
#define I_SPR		1024
#define SPR			(&regs[I_SPR])
// for paired singles
#define I_GQR		(I_SPR + 912)
#define GQR			(&regs[I_GQR])
#define GQR0		(GQR[0])
#define GQR1		(GQR[1])
#define GQR2		(GQR[2])
#define GQR3		(GQR[3])
#define GQR4		(GQR[4])
#define GQR5		(GQR[5])
#define GQR6		(GQR[6])
#define GQR7		(GQR[7])
// program counter
#define I_PC		4095
#define PC			(regs[I_PC])
// instruction counter
#define I_IC		4094
#define IC			(regs[I_IC])

// stackpointer
#define I_SP		1
#define SP			(regs[I_SP])
#define I_SDA1	2
#define I_SDA2	13
#define SDA1		(regs[I_SDA1])
#define SDA2		(regs[I_SDA2])
// register specifing carries and overflows for integer operations
#define I_XER		(I_SPR + 1)
#define XER			(regs[I_XER])
// link register
#define I_LR		(I_SPR + 8)
#define LR			(regs[I_LR])
// count register
#define I_CTR		(I_SPR + 9)
#define CTR			(regs[I_CTR])
// decrement register
#define I_DEC		(I_SPR + 22)
#define DEC			(regs[I_DEC])
// machine status save/restore register 0 and 1
#define I_SRR0	(I_SPR + 26)
#define I_SRR1	(I_SPR + 27)
#define SRR0		(regs[I_SRR0])
#define SRR1		(regs[I_SRR1])
// hardware implementation registers
#define I_HID2	(I_SPR + 920)
#define I_HID0	(I_SPR + 1008)
#define I_HID1	(I_SPR + 1009)
#define HID2		(regs[I_HID2])
#define HID0		(regs[I_HID0])
#define HID1		(regs[I_HID1])

#define I_WPAR	(I_SPR + 921)
#define WPAR		(regs[I_WPAR])
#define I_DMAU	(I_SPR + 922)
#define I_DMAL	(I_SPR + 923)
#define DMAU		(regs[I_DMAU])
#define DMAL		(regs[I_DMAL])

// floating point registers (access as fp and binary)
#define FPR			((double *)fpregs)
#define FBR			((__u64  *)fpregs)

// opcode decoding
#define OPC		 	opcode
#define RD			((OPC >> 21) &       0x1f)
#define RS			RD
#define RA			((OPC >> 16) &       0x1f)
#define RB			((OPC >> 11) &       0x1f)
#define RC			((OPC >>  6) &       0x1f)
#define UIMM		((__u16) OPC)
#define SIMM		((__s16) OPC)
#define OE			((OPC >> 10) &       0x01)
#define Rc			(OPC         &       0x01)
#define LK			Rc
#define AA			((OPC >>  1) &       0x01)
#define LI			(OPC         &  0x3fffffc)
#define BO			((OPC >> 21) &       0x1f)
#define BI			((OPC >> 16) &       0x1f)
#define BD			(OPC         &     0xfffc)
#define CRFD		((OPC >> 23) &       0x07)
#define L				((OPC >> 21) &       0x01)
#define CRBD		RD
#define CRBA		RA
#define CRBB		RB
#define CRFS		((OPC >> 18) &       0x07)
#define CRM			((OPC >> 12) &       0xff)
#define SH			RB
#define MB			((OPC >>  6) &       0x1f)
#define ME			((OPC >>  1) &       0x1f)
#define FM			((OPC >> 17) &       0xff)
#define IMM			((OPC >> 12) &       0x0f)

// quick register access
#define RRD			(GPR[RD])
#define RRS			(GPR[RS])
#define RRA			(GPR[RA])
#define RRB			(GPR[RB])
#define RRC			(GPR[RC])
#define NB			RB
#define RNB			(GPR[NB])
#define FRD			(FPR[RD])
#define FRS			(FPR[RS])
#define FRA			(FPR[RA])
#define FRB			(FPR[RB])
#define FRC			(FPR[RC])
#define FBD			(FBR[RD])
#define FBS			(FBR[RS])
#define FBA 		(FBR[RA])
#define FBB 		(FBR[RB])
#define FBC 		(FBR[RC])

// paired singles
#define PS						((float *)fpregs)
#define PSB						((__u32 *)fpregs)
#define PS0						((float *)fpregs)
#define PS1						((float *)&fpregs[32])
#define PSB0					((__u32 *)fpregs)
#define PSB1					((__u32 *)&fpregs[32])
#define PSINDEX(ps,i)	((i << 1) + (1-ps))
#define PS0D					(PS[PSINDEX (0, RD)])
#define PS1D					(PS[PSINDEX (1, RD)])
#define PS0A					(PS[PSINDEX (0, RA)])
#define PS1A					(PS[PSINDEX (1, RA)])
#define PS0B					(PS[PSINDEX (0, RB)])
#define PS1B					(PS[PSINDEX (1, RB)])
#define PS0C					(PS[PSINDEX (0, RC)])
#define PS1C					(PS[PSINDEX (1, RC)])
#define PSB0D					(PSB[PSINDEX (0, RD)])
#define PSB1D					(PSB[PSINDEX (1, RD)])
#define PSB0A					(PSB[PSINDEX (0, RA)])
#define PSB1A					(PSB[PSINDEX (1, RA)])
#define PSB0B					(PSB[PSINDEX (0, RB)])
#define PSB1B					(PSB[PSINDEX (1, RB)])
#define PSB0C					(PSB[PSINDEX (0, RC)])
#define PSB1C					(PSB[PSINDEX (1, RC)])
#define PSIMM					(EXTS (12, ((OPC) & 0xfff)))
#define PSI						((OPC >> 12) & 0x07)
#define PSXI					((OPC >>  7) & 0x07)
#define PSW						(OPC & 0x8000)
#define PSXW					(OPC & 0x0400)
#define LD_SCALE(n)		((GQR[n] >> 24) & 0x3f)
#define LD_TYPE(n)		((GQR[n] >> 16) & 0x07)
#define ST_SCALE(n)		((GQR[n] >>  8) & 0x3f)
#define ST_TYPE(n)		((GQR[n]      ) & 0x07)


#define BRANCH_CTR_OK(bo)		(BIT_IS_SET ((bo), 5, 2) || (((CTR != 0) ^ BIT ((bo), 5, 3))))
#define BRANCH_COND_OK(bo) 	(BIT_IS_SET ((bo), 5, 0) || (BIT (CR, 32, BI) == BIT ((bo), 5, 1)))


// condition register bits
// lower than, greater than, equal and summary overflow (copy of XER_SO)
#define CR_LT_BIT		(3-0)
#define CR_GT_BIT		(3-1)
#define CR_EQ_BIT		(3-2)
#define CR_SO_BIT		(3-3)
#define CR_LT				(1 << CR_LT_BIT)
#define CR_GT				(1 << CR_GT_BIT)
#define CR_EQ				(1 << CR_EQ_BIT)
#define CR_SO				(1 << CR_SO_BIT)


#define SET_FIELD(X,n,Y)\
	({\
			__u32 _n = (n);\
			(X = (X &~ (0x0f << ((7-(_n)) << 2))) | ((Y) << ((7-(_n)) << 2)));\
	})
#define CLEAR_FIELD(X,n)	(X = (X &~ (0x0f << ((7-(n)) << 2))))
#define GET_FIELD(X,n)		((X >> (32 - (((n) + 1) << 2))) & 0x0f)


#define CALC_CR(n,X)\
	({\
			__s32 _X = (X);\
			SET_FIELD (CR, n, ((IS_XER_SO) ? CR_SO : 0) | (((_X) < 0) ? CR_LT : 0) | (((_X) > 0) ? CR_GT : 0) | (((_X) == 0) ? CR_EQ : 0));\
	})

// unsigned
#define CALC_CRL(n,X,Y)\
	({\
			__u32 _X = (X);\
			__u32 _Y = (Y);\
			SET_FIELD (CR, n, ((IS_XER_SO) ? CR_SO : 0) | ((_X < _Y) ? CR_LT : 0) | ((_X > _Y) ? CR_GT : 0) | ((_X == _Y) ? CR_EQ : 0));\
	})

// rewrite in asm
#define CALC_CR0(X)				(CALC_CR (0,X))
#define SET_CR0(X)				(SET_FIELD (CR, 0, X))
#define GET_CR0						(GET_FIELD (CR, 0))

#define IS_NAN(X)					((((X) & 0x000fffffffffffffULL) != 0) && (((X) & 0x7ff0000000000000ULL) == 0x7ff0000000000000ULL))
#define CALC_CRF(n,Xf,Xb,Yf,Yb)\
	({\
			double _X = (Xf);\
			double _Y = (Yf);\
			if (IS_NAN (Xb) || IS_NAN (Yb))\
				SET_FIELD (CR, n, 1);\
			else if (_X < _Y)\
				SET_FIELD (CR, n, 8);\
			else if (_X > _Y)\
				SET_FIELD (CR, n, 4);\
			else\
				SET_FIELD (CR, n, 2);\
	})

#define IS_NANS(X)			((((X) & 0x007fffff) != 0) && (((X) & 0x7f800000) == 0x7f800000))
#define CALC_CRFS(n,Xf,Xb,Yf,Yb)\
	({\
			float _X = (Xf);\
			float _Y = (Yf);\
			if (IS_NANS (Xb) || IS_NANS (Yb))\
				SET_FIELD (CR, n, 1);\
			else if (_X < _Y)\
				SET_FIELD (CR, n, 8);\
			else if (_X > _Y)\
				SET_FIELD (CR, n, 4);\
			else\
				SET_FIELD (CR, n, 2);\
	})

#define SQRT(X)						(sqrt (X))


// XER bits
// summary overflow, overflow and carry
// SO - set whenever an instruction sets the OV bit (except mtspr)
//      once set it remains set until cleared by mtspr or mcrxr
// OV - add, sub and neg (with OE = 1) set it if the carry out of the msb
//      is not equal to the carry out of the msb + 1, and clear it otherwise
//      mul and div set it if the result cannot be represented with 32 bits
#define XER_SO_BIT		((2-0) + 29)
#define XER_OV_BIT		((2-1) + 29)
#define XER_CA_BIT		((2-2) + 29)
#define XER_SO				(1 << XER_SO_BIT)
#define XER_OV				(1 << XER_OV_BIT)
#define XER_CA				(1 << XER_CA_BIT)
// XER byte count
#define XER_BC				(XER & 0x7f)

#define IS_XER_CA		 			((XER & XER_CA) != 0)
#define SET_XER_CA				(XER |=  XER_CA)
#define CLEAR_XER_CA			(XER &= ~XER_CA)
#define CHANGE_XER_CA(X)	((X)? SET_XER_CA : CLEAR_XER_CA)

#define IS_XER_OV		 			((XER & XER_OV) != 0)
#define SET_XER_OV				(XER |=  XER_OV)
#define CLEAR_XER_OV			(XER &= ~XER_OV)

#define IS_XER_SO		 			((XER & XER_SO) != 0)
#define SET_XER_SO				(XER |=  XER_SO)
#define CLEAR_XER_SO			(XER &= ~XER_SO)

#define SET_XER_OV_SO			(XER |=  XER_OV | XER_SO)

#define CALC_OVERFLOW(X,Y) 	(((X) > 0) && ((Y) > INT_MAX - (X)))
#define CALC_UNDERFLOW(X,Y) (((X) < 0) && ((Y) < INT_MIN - (X)))

// carry
#define CALC_XER_CA(X,Y)			(((X) > UINT_MAX - (Y)) ? SET_XER_CA : CLEAR_XER_CA)
#define CALC_XER_CA_E(X,Y,c)	((!c) ? CLEAR_XER_CA : (((X)+(Y) == UINT_MAX)? SET_XER_CA : CLEAR_XER_CA))
// overflow
#define CALC_XER_OV(X,Y)			((((X) > 0) && ((Y) > INT_MAX - (X))) ? SET_XER_OV_SO : CLEAR_XER_OV)
#define CALC_XER_OV_E(X,Y,c)	((((X) > 0) && ((Y) > INT_MAX - (X))) ? SET_XER_OV_SO : ((c) ? CALC_XER_OV ((X)+(Y), c) : CLEAR_XER_OV))
// underflow
#define CALC_XER_OV_U(X,Y)		((((X) < 0) && ((Y) < INT_MIN - (X))) ? SET_XER_OV_SO : CLEAR_XER_OV)
#define CALC_XER_OV_UE(X,Y,c)	((((X) < 0) && ((Y) < INT_MIN - (X))) ? SET_XER_OV_SO : ((c) ? CALC_XER_OV ((X)+(Y), c) : CLEAR_XER_OV))

// 0-3 bits of XER
#define GET_XER_03				((XER >> (31 - 3)) & 0x07)
#define SET_XER_03(X)			(XER = (XER &~ (0x0f << (31-3))) | ((X) << (31-3)))

#define GENMASK(X,Y)\
	({\
			__u32 _X = (X);\
			__u32 _Y = (Y);\
			((_X <= _Y) ? \
			((0xffffffff >> (_X)) ^ (((_Y) < 31)? (0xffffffff >> ((_Y) + 1)) : 0)) :\
			(~((0xffffffff >> (_X)) ^ (((_Y) < 31)? (0xffffffff >> ((_Y) + 1)) : 0)))) ;\
	})
#define MASK(X,Y)			(mask[X][Y])

#ifndef ROTL
# define ROTL(X,n)\
	({\
			__u32 _X = (X);\
			__u32 _n = (n);\
			((_X << _n) | (_X >> (32 - _n)));\
	})
#endif


#define EXTS(n,X)					((__s32)((X) << (32 - (n))) >> (32 - (n)))
#define EXTS8(X)					((__u32)(__s32)(__s8)(__u8) X)
#define EXTS16(X)					((__u32)(__s32)(__s16)(__u16) X)

#define FPSCR_FEX					(1 << 30)
#define FPSCR_VX					(1 << 29)

#define MSR_POW						(1 << (31-13))
#define MSR_ILE						(1 << (31-15))
#define MSR_EE						(1 << (31-16))
#define MSR_PR						(1 << (31-17))
#define MSR_FP						(1 << (31-18))
#define MSR_ME						(1 << (31-19))
#define MSR_FE0						(1 << (31-20))
#define MSR_SE						(1 << (31-21))
#define MSR_BE						(1 << (31-22))
#define MSR_FE1						(1 << (31-23))
#define MSR_IP						(1 << (31-25))
#define MSR_IR						(1 << (31-26))
#define MSR_DR						(1 << (31-27))
#define MSR_RI						(1 << (31-30))
#define MSR_LE						(1 << (31-31))

// exceptions
#define EXCEPTION_SYSTEM_RESET					0x80000100
#define EXCEPTION_MACHINE_CHECK					0x80000200
#define EXCEPTION_DSI										0x80000300
#define EXCEPTION_ISI										0x80000400
#define EXCEPTION_EXTERNAL							0x80000500
#define EXCEPTION_ALIGNMENT							0x80000600
#define EXCEPTION_PROGRAM								0x80000700
#define EXCEPTION_FP_UNAVAILABLE				0x80000800
#define EXCEPTION_DECREMENTER						0x80000900
#define EXCEPTION_SYSTEM_CALL						0x80000c00
#define EXCEPTION_TRACE									0x80000d00
#define EXCEPTION_PERFORMANCE_MONITOR		0x80000f00
#define EXCEPTION_IABR									0x80001300
#define EXCEPTION_RESERVED							0x80001400
#define EXCEPTION_THERMAL								0x80001700


#define QLOAD(ea,offset,type,scale,dest)		\
	({\
		switch (type)\
		{\
			case 0:\
				*(dest) = MEM_RSINGLE_F (ea + (offset << 2));\
				break;\
			case 4:\
				*(dest) = (double) MEM_RBYTE (ea + offset);\
				break;\
			case 5:\
				*(dest) = (double) MEM_RWORD (ea + (offset << 1));\
				break;\
			case 6:\
				*(dest) = (double) (__s8) MEM_RBYTE (ea + offset);\
				break;\
			case 7:\
				*(dest) = (double) (__s16) MEM_RWORD (ea + (offset << 1));\
				break;\
			default:\
				DEBUG (EVENT_EMAJOR, ".cpu: unhandled qload: (%d,%d)", type, scale);\
		}\
	})


#define QSTORE(ea,offset,type,scale,dest)		\
	({\
		switch (type)\
		{\
			case 0:\
				MEM_WSINGLE (ea + (offset << 2), dest);\
				break;\
			case 4:\
				MEM_WBYTE (ea + offset, (__u8) dest);\
				break;\
			case 5:\
				MEM_WWORD (ea + (offset << 1), (__u16) dest);\
				break;\
			case 6:\
				MEM_WBYTE (ea + offset, (__s8) dest);\
				break;\
			case 7:\
				MEM_WWORD (ea + (offset << 1), (__s16) dest);\
				break;\
			default:\
				DEBUG (EVENT_EMAJOR, ".cpu: unhandled qstore: (%d,%d)", type, scale);\
		}\
	})


void cpu_init (void);
void cpu_execute (void);
void cpu_exception (__u32 ex);


#endif // __CPU_H
